"""Types used in SHV RPC method description."""

from __future__ import annotations

import collections.abc
import dataclasses
import enum
import functools
import typing

from .rpcparam import SHVGetKey, shvget, shvgett
from .value import SHVMapType, SHVType, is_shvmap


class RpcMethodFlags(enum.IntFlag):
    """Flags assigned to the SHV RPC methods."""

    NOT_CALLABLE = 1 << 0
    GETTER = 1 << 1
    SETTER = 1 << 2
    LARGE_RESULT_HINT = 1 << 3
    NOT_IDEMPOTENT = 1 << 4
    USER_ID_REQUIRED = 1 << 5


class RpcMethodAccess(enum.IntEnum):
    """Method access level."""

    BROWSE = 1
    READ = 8
    WRITE = 16
    COMMAND = 24
    CONFIG = 32
    SERVICE = 40
    SUPER_SERVICE = 48
    DEVEL = 56
    ADMIN = 63

    @classmethod
    @functools.cache
    def strmap(cls) -> dict[str, RpcMethodAccess]:
        """Map from string to this enum."""
        return {
            "bws": cls.BROWSE,
            "rd": cls.READ,
            "wr": cls.WRITE,
            "cmd": cls.COMMAND,
            "cfg": cls.CONFIG,
            "srv": cls.SERVICE,
            "ssrv": cls.SUPER_SERVICE,
            "dev": cls.DEVEL,
            "su": cls.ADMIN,
        }

    @classmethod
    @functools.cache
    def strrmap(cls) -> dict[int, str]:
        """Map from this enum to string."""
        return {v.value: k for k, v in cls.strmap().items()}

    def tostr(self) -> str:
        """Convert to string representation."""
        return self.strrmap().get(self.value, "bws")

    @classmethod
    def fromstr(cls, access: str) -> RpcMethodAccess:
        """Convert to string representation."""
        return cls.strmap().get(access, cls.BROWSE)


@dataclasses.dataclass
class RpcMethodDesc:
    """Description of the SHV RPC method.

    This is implemented as :func:`dataclasses.dataclass`.

    :param name: Name of the method.
    :param flags: Flags assigned to the method.
    :param param: Parameter type that should be provided to the method.
    :param result: Result type that is provided by the method.
    :param access: Minimal granted access level for this method.
    :param signals: Mapping of signal name to data type they cary. These are
      signals emited by this method.
    :param extra: Additional fields to be provied in some cases. The most common
      one is ``"description"`` with method description.
    """

    class Key(enum.IntEnum):
        """Key in the description IMap."""

        NAME = 1
        FLAGS = 2
        PARAM = 3
        RESULT = 4
        ACCESS = 5
        SIGNALS = 6
        EXTRA = 63

    name: str
    flags: RpcMethodFlags = dataclasses.field(default=RpcMethodFlags(0))
    param: str = "Null"
    result: str = "Null"
    access: RpcMethodAccess = RpcMethodAccess.BROWSE
    signals: dict[str, str] = dataclasses.field(default_factory=dict)
    extra: dict[str, SHVType] = dataclasses.field(default_factory=dict)

    def to_shv(self, extra: bool = False) -> SHVType:
        """Convert to SHV RPC representation."""
        res: dict[int, SHVType] = {
            self.Key.NAME: self.name,
            self.Key.FLAGS: self.flags,
        }
        if self.param != "Null":
            res[self.Key.PARAM] = self.param
        if self.result != "Null":
            res[self.Key.RESULT] = self.result
        res[self.Key.ACCESS] = self.access
        if self.signals:
            res[self.Key.SIGNALS] = {
                k: None if v == self.result else v for k, v in self.signals.items()
            }
        if self.extra:
            res[self.Key.EXTRA] = self.extra
        return res

    @classmethod
    def from_shv(cls, value: SHVType) -> RpcMethodDesc:
        """Create from SHV RPC representation."""
        if not isinstance(value, collections.abc.Mapping):
            raise ValueError(f"Expected Map but got {value!r}.")
        raccess = shvget(value, SHVGetKey("access", cls.Key.ACCESS), cls.access)
        rsignals = shvget(
            value, SHVGetKey("source", cls.Key.SIGNALS), typing.cast(SHVMapType, {})
        )
        rextra = shvget(value, cls.Key.EXTRA, typing.cast(SHVMapType, {}))
        result: str = shvgett(value, cls.Key.RESULT, str, cls.result)
        return cls(
            name=shvgett(value, SHVGetKey("name", cls.Key.NAME), str, "UNSPECIFIED"),
            flags=RpcMethodFlags(
                shvgett(value, SHVGetKey("flags", cls.Key.FLAGS), int, cls.flags)
            ),
            param=shvgett(value, cls.Key.PARAM, str, cls.param),
            result=result,
            access=RpcMethodAccess(raccess)
            if isinstance(raccess, int)
            else RpcMethodAccess.fromstr(raccess)
            if isinstance(raccess, str)
            else cls.access,
            signals={
                k: (result if v is None else v)
                for k, v in rsignals.items()
                if v is None or isinstance(v, str)
            }
            if is_shvmap(rsignals)
            else {},
            extra=dict(rextra) if is_shvmap(rextra) else {},
        )

    @classmethod
    def getter(
        cls,
        name: str = "get",
        param: str = "Int",
        result: str = "Any",
        access: RpcMethodAccess = RpcMethodAccess.READ,
        signal: bool | str = False,
        description: str = "",
    ) -> RpcMethodDesc:
        """Create getter method description.

        :param name: Name of the method.
        :param param: Type of the parameter this getter expects.
        :param result: Type of the result this getter provides.
        :param access: Minimal granted access level for this getter.
        :param signal: Allows specifying property change signal. You can specify
          `True` to get default `"chng"` signal or you can specify custom signal
          name (should end with *chng*).
        :param description: Short description of the value.
        """
        return cls(
            name,
            RpcMethodFlags.GETTER,
            param,
            result,
            access,
            {"chng" if signal is True else signal: result} if signal else {},
            {"description": description} if description else {},
        )

    @classmethod
    def setter(
        cls,
        name: str = "set",
        param: str = "Any",
        result: str = "Null",
        access: RpcMethodAccess = RpcMethodAccess.WRITE,
        description: str = "",
    ) -> RpcMethodDesc:
        """Create setter method description.

        :param name: Name of the method.
        :param param: Type of the parameter this setter expects.
        :param result: Type of the result this setter provides.
        :param access: Minimal granted access level for this setter.
        :param description: Short description of the value.
        """
        return cls(
            name,
            RpcMethodFlags.SETTER,
            param,
            result,
            access,
            extra={"description": description} if description else {},
        )

    @classmethod
    @functools.lru_cache(maxsize=1)
    def stddir(cls) -> RpcMethodDesc:
        """Get description of standard 'dir' method."""
        return cls("dir", param="idir", result="odir")

    @classmethod
    @functools.lru_cache(maxsize=1)
    def stdls(cls) -> RpcMethodDesc:
        """Get description of standard 'ls' method."""
        return cls("ls", param="ils", result="ols", signals={"lsmod": "olsmod"})



"""Just some utilities we use in broker."""

from __future__ import annotations

import abc
import collections.abc
import itertools
import typing


class Comparable(typing.Protocol):
    """Protocol for annotating comparable types."""

    @abc.abstractmethod
    def __gt__(self: T, other: T) -> bool:
        pass

    @abc.abstractmethod
    def __lt__(self: T, other: T) -> bool:
        pass


T = typing.TypeVar("T", bound=Comparable)


def nmin(*vals: collections.abc.Iterator[T | None] | T | None) -> T | None:
    """Get minimum or ``None`` if all values were ``None``."""
    res: T | None = None
    itr = (v if isinstance(v, collections.abc.Iterator) else iter([v]) for v in vals)
    for val in itertools.chain(*itr):
        if val is not None and (res is None or res < val):
            res = val
    return res


def nmax(*vals: collections.abc.Iterator[T | None] | T | None) -> T | None:
    """Get maximum or ``None`` if all values were ``None``."""
    res: T | None = None
    itr = (v if isinstance(v, collections.abc.Iterator) else iter([v]) for v in vals)
    for val in itertools.chain(*itr):
        if val is not None and (res is None or res > val):
            res = val
    return res



"""Abstract definition of the broker configuration."""

from __future__ import annotations

import abc
import collections
import collections.abc

from ..rpclogin import RpcLogin
from ..rpcmethod import RpcMethodAccess
from ..rpcurl import RpcUrl


class RpcBrokerRoleABC(abc.ABC):
    """Role used to control the peer's access and setup."""

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Name of the role used to identify it externally."""

    def mount_point(  # noqa PLR6301
        self, existing: collections.abc.Set[str] = frozenset()
    ) -> str | None:
        """Provide mount point for this role.

        Note that root node, ``.app``, and ``.broker`` are never allowed to
        be mounted and thus should bever be returned.

        :param existing: Set of all existing mount points. They are provided
          to avoid these when mount points are being generated.
        :return: Iterator that generates possible mount points. Iterator that
          provides no item or just one is most common but generated mount points
          can be used as well.
        """
        return None

    def initial_subscriptions(self) -> collections.abc.Iterator[str]:  # noqa PLR6301
        """Iterate over subscription the peer should be initialized with.

        These subscription should be inserted as the initial set of
        subscriptions (TTL won't be applied on them).

        :return: Iterator over RPC RI for subscriptions.
        """
        return iter([])

    def access_level(  # noqa PLR6301
        self, path: str, method: str
    ) -> RpcMethodAccess | None:
        """Deduce the access level (if any) for given method.

        Note that all users have implicit browse access to the root node,
        ``.app``, ``.broker``, and ``.broker/currentClient`` that doesn't
        have to be covered by this function.

        :param path: SHV Path of the method.
        :param method: SHV Method name.
        :return: Access level for this method or ``None`` in case of no access.
        """
        return None


class RpcBrokerConfigABC(abc.ABC):
    """Abstract base for the SHV Broker configuration."""

    @property
    def name(self) -> str:
        """Name of the broker used to identify broker in UserID.

        The default is empty string and in such case broker name is not
        considered.
        """
        return ""

    def listens(self) -> collections.abc.Iterator[RpcUrl]:  # noqa PLR6301
        """Iterate over URLs where Broker should listen to."""
        return iter([])

    def connections(self) -> collections.abc.Iterator[tuple[RpcUrl, RpcBrokerRoleABC]]:  # noqa PLR6301
        """Iterate over URLs and their setup where Broker should connect to."""
        return iter([])

    @abc.abstractmethod
    def login(self, login: RpcLogin, nonce: str) -> RpcBrokerRoleABC | None:
        """Check the login and provide role if login is correct."""



"""Check implementation of SimpleDevice."""

import datetime

import pytest

from shv import (
    SHV_VERSION_MAJOR,
    SHV_VERSION_MINOR,
    VERSION,
    RpcMethodAccess,
    RpcMethodDesc,
    RpcMethodNotFoundError,
    SimpleDevice,
    shvmeta,
)


class Device(SimpleDevice):
    DEVICE_NAME = "testdev"
    DEVICE_VERSION = "0.0.x"


@pytest.fixture(name="device")
async def fixture_device(shvbroker, url_test_device):
    """Run device and provide instance to access it."""
    device = await Device.connect(url_test_device)
    yield device
    await device.disconnect()


@pytest.mark.parametrize(
    "path,method,result",
    (
        ("test/device/.app", "shvVersionMajor", SHV_VERSION_MAJOR),
        ("test/device/.app", "shvVersionMinor", SHV_VERSION_MINOR),
        ("test/device/.app", "name", "pyshv-device"),
        ("test/device/.app", "version", VERSION),
        ("test/device/.app", "ping", None),
        ("test/device/.device", "name", "testdev"),
        ("test/device/.device", "version", "0.0.x"),
        ("test/device/.device", "serialNumber", None),
    ),
)
async def test_call(client, device, path, method, result):
    """Check that we can call various methods using blocking call."""
    res = await client.call(path, method)
    assert res == result
    assert shvmeta(res) == shvmeta(result)


async def test_call_date(client, device):
    """Check that we can call various methods using blocking call."""
    res = await client.call("test/device/.app", "date")
    assert isinstance(res, datetime.datetime)
    assert res.tzinfo is not None
    diff = datetime.datetime.now().astimezone() - res
    assert datetime.timedelta() <= diff < datetime.timedelta(seconds=1)


async def test_invalid_call(client):
    with pytest.raises(RpcMethodNotFoundError):
        await client.call(".app", "someInvalid")


@pytest.mark.parametrize(
    "path,result",
    (
        ("test/device", [".app", ".device"]),
        ("test/device/.app", []),
        ("test/device/.device", []),
    ),
)
async def test_ls(client, device, path, result):
    """Verify that we can use ls method."""
    res = await client.ls(path)
    assert res == result


@pytest.mark.parametrize(
    "path,result",
    (
        (
            "test/device/.device",
            [
                RpcMethodDesc.stddir(),
                RpcMethodDesc.stdls(),
                RpcMethodDesc.getter(
                    "name", "Null", "String", access=RpcMethodAccess.BROWSE
                ),
                RpcMethodDesc.getter(
                    "version", "Null", "String", access=RpcMethodAccess.BROWSE
                ),
                RpcMethodDesc.getter(
                    "serialNumber",
                    "Null",
                    "OptionalString",
                    access=RpcMethodAccess.BROWSE,
                ),
            ],
        ),
    ),
)
async def test_dir(client, device, path, result):
    """Verify that we can use dir method."""
    res = await client.dir(path)
    assert res == result



"""RPC client specific functions to login to the RPC Broker."""

from __future__ import annotations

import copy
import dataclasses
import enum
import functools
import getpass
import hashlib
import logging
import typing

from .rpcerrors import RpcInvalidParamError
from .rpcparam import shvgett
from .value import SHVMapType, SHVType, is_shvmap

logger = logging.getLogger(__name__)

T = typing.TypeVar("T")


@functools.lru_cache
def _get_user() -> str:
    # getuser fails if there is no account assigned to the UID in the system
    try:
        return getpass.getuser()
    except KeyError:  # pragma: no cover
        return "nobody"


class RpcLoginType(enum.Enum):
    """Enum specifying which login type should be used.

    The string values are the exact string representation used in SHV RPC
    protocol identifying these login types.
    """

    PLAIN = "PLAIN"
    """Plain login format should be used."""
    SHA1 = "SHA1"
    """Use hash algorithm SHA1 (preferred and common default)."""


@dataclasses.dataclass
class RpcLogin:
    """SHV RPC Broker login info.

    This is all in one login info for user.
    """

    username: str = dataclasses.field(default=_get_user())
    """User name used to login to the remote server."""
    password: str = ""
    """Password used to login to the server."""
    login_type: RpcLoginType = RpcLoginType.PLAIN
    """Type of the login to be used (specifies format of the password)."""
    options: dict[str, SHVType] = dataclasses.field(default_factory=dict)
    """Additional options passed with login to the RPC Broker to configure session."""

    opt_device_id: dataclasses.InitVar[str | None] = None
    opt_device_mount_point: dataclasses.InitVar[str | None] = None

    def __post_init__(
        self, opt_device_id: str | None, opt_device_mount_point: str | None
    ) -> None:
        if opt_device_id is not None:
            self.device_id = opt_device_id
        if opt_device_mount_point is not None:
            self.device_mount_point = opt_device_mount_point

    @property
    def device_id(self) -> str | None:
        """Device identifier sent to the server with login."""
        return self.__dictget(self.options, str, "device", "deviceId")

    @device_id.setter
    def device_id(self, value: str | None) -> None:
        self.__dictmerge({"device": {"deviceId": value}}, self.options)

    @property
    def device_mount_point(self) -> str | None:
        """Request for mounting of connected device to the specified mount point."""
        return self.__dictget(self.options, str, "device", "mountPoint")

    @device_mount_point.setter
    def device_mount_point(self, value: str | None) -> None:
        self.__dictmerge({"device": {"mountPoint": value}}, self.options)

    @property
    def idle_timeout(self) -> int | None:
        """Request for specific setting of the idle timeout on the broker."""
        return self.__dictget(self.options, int, "idleWatchDogTimeOut")

    @idle_timeout.setter
    def idle_timeout(self, value: int | None) -> None:
        self.__dictmerge({"idleWatchDogTimeOut": value}, self.options)

    def extend_options(self, options: dict[str, SHVType]) -> None:
        """Merge passed options with curent ones.

        This is handy if you have your options and only need to fill standard
        ones provided by this class.

        :param options: Dictionary with items used to extend options.
        """
        self.__dictmerge(options, self.options)

    def validate_password(
        self, password: str, nonce: str, login_type: RpcLoginType = RpcLoginType.PLAIN
    ) -> bool:
        """Validate this login against given password and nonce.

        The arguments passed to this method are for the expected password that
        user should provide. The password can be either in plain text or hashed
        with SHA1.

        :param password: The reference password.
        :param nonce: Nonce used for SHA1 login.
        :param login_type: The format of the password.
        """
        rpass = self.password
        match login_type, self.login_type:
            case RpcLoginType.PLAIN, RpcLoginType.PLAIN:
                pass
            case RpcLoginType.PLAIN, RpcLoginType.SHA1:
                password = hashlib.sha1(password.encode("utf-8")).hexdigest()  # noqa PLR6301
                login_type = RpcLoginType.SHA1
            case RpcLoginType.SHA1, RpcLoginType.PLAIN:
                rpass = hashlib.sha1(  # noqa PLR6301
                    nonce.encode("utf-8")
                    + hashlib.sha1(rpass.encode("utf-8")).hexdigest().encode("utf-8")  # noqa PLR6301
                ).hexdigest()
            case RpcLoginType.SHA1, RpcLoginType.SHA1:
                pass
            case _, _:  # pragma: no cover
                return False
        match login_type:
            case RpcLoginType.PLAIN:
                return rpass == password
            case RpcLoginType.SHA1:
                m = hashlib.sha1()  # noqa PLR6301
                m.update(nonce.encode("utf-8"))
                m.update(password.encode("utf-8"))  # password must be SHA1
                return rpass == m.hexdigest()
            case _:  # pragma: no cover
                raise NotImplementedError

    @classmethod
    def __dictget(cls, src: SHVType, tp: type[T], *key: str) -> T | None:
        if not key:
            return src if isinstance(src, tp) else None
        if not is_shvmap(src) or key[0] not in src:
            return None
        return cls.__dictget(src[key[0]], tp, *key[1:])

    @classmethod
    def __dictmerge(
        cls, src: SHVMapType, dest: dict[str, SHVType]
    ) -> dict[str, SHVType]:
        for k, v in src.items():
            if (
                is_shvmap(v)
                and isinstance((destk := dest.get(k, None)), dict)
                and all(isinstance(k, str) for k in destk)
            ):
                cls.__dictmerge(v, destk)
                if not destk:
                    dest.pop(k, None)
            elif cls.__notnone(v):
                dest[k] = copy.deepcopy(v)
            else:
                dest.pop(k, None)
        return dest

    @classmethod
    def __notnone(cls, value: SHVType) -> bool:
        if not is_shvmap(value):
            return value is not None
        return all(cls.__notnone(v) for v in value.values())

    def to_shv(
        self,
        nonce: str,
        custom_options: SHVMapType | None = None,
        trusted: bool = False,
    ) -> SHVType:
        """RPC Login parameter in the parameter format for login method.

        :param nonce: The string with random characters returned from hello
          method call that is used for SHA1 login.
        :param custom_options: These are additional options to be added to the
          login options.
        :param trusted: Specifies if the transport can be trusted. On untrusted
          transport layers the `SHA1` password is used even if this login
          specifies `PLAIN`.
        :return: Parameters to be passed to login method call.
        """
        password = self.password
        login_type = self.login_type
        if self.login_type is RpcLoginType.PLAIN and not trusted:
            login_type = RpcLoginType.SHA1
            password = hashlib.sha1(self.password.encode("utf-8")).hexdigest()  # noqa S324
        if login_type is RpcLoginType.SHA1:
            m = hashlib.sha1()  # noqa S324
            m.update(nonce.encode("utf-8"))
            m.update((password or "").encode("utf-8"))
            password = m.hexdigest()
        return {
            "login": {
                "password": password,
                "type": login_type.value,
                "user": self.username,
            },
            "options": typing.cast(
                SHVType,
                self.__dictmerge(
                    self.options, copy.deepcopy(dict(custom_options or {}))
                ),
            ),
        }

    @classmethod
    def from_shv(cls, value: SHVType) -> RpcLogin:
        """Create from SHV RPC login method parameter.

        :param value: The value that was received as login method parameter.
        :return: The :class:`RpcLogin` object.
        :raise RpcInvalidParamError: If parameter is invalid in some way.
        """
        if not is_shvmap(value):
            raise RpcInvalidParamError("Expected Map.")
        username = shvgett(value, ("login", "user"), str, "")
        password = shvgett(value, ("login", "password"), str, "")
        login_type = RpcLoginType(
            shvgett(
                value,
                ("login", "type"),
                str,
                RpcLoginType.SHA1.value
                if len(password) == 40
                else RpcLoginType.PLAIN.value,
            )
        )
        options = value.get("options")
        if not is_shvmap(options):
            options = {}
        return cls(username, password, login_type, dict(options))
